@@ -1,6 +1,6 @@
 /*
  * version.h
- * Copyright (C) 1998-2001 A.J. van Os; Released under GPL
+ * Copyright (C) 1998-2004 A.J. van Os; Released under GNU GPL
  *
  * Description:
  * Version and release information
@@ -13,12 +13,20 @@
 #define PURPOSESTRING	"Display MS-Word files"
 
 #if defined(__riscos)
-#define AUTHORSTRING	" 1998-2001 Adri van Os"
+#define AUTHORSTRING	" 1998-2004 Adri van Os"
 #else
-#define AUTHORSTRING	"(C) 1998-2001 Adri van Os"
+#define AUTHORSTRING	"(C) 1998-2004 Adri van Os"
 #endif /* __riscos */
 
-#define VERSIONSTRING	"0.32  (05 Oct 2001)"
+#define VERSIONSTRING	"0.36.1  (09 Dec 2004)"
+
+#if defined(__dos)
+#if defined(__DJGPP__)
+#define VERSIONSTRING2	" # 32-bit Protected Mode"
+#else
+#define VERSIONSTRING2	" # 16-bit Real Mode"
+#endif /* __DJGPP__ */
+#endif /* __dos */
 
 #if defined(DEBUG)
 #define STATUSSTRING	"DEBUG version"
